=======================
{{ config }}
{% for result in llm_results %}
Q: {{ result.question }}
A: {{ result.answer }}
Source: {% for key, value in dict_crosstab(result.source_documents).items() %}
    document: {{ key }}, page: {{ value }} {% endfor %}
{% endfor %}=======================